import csv
import statistics


class Utils:

    def __init__(self, rules):
        self.rules = rules
        self.num_rules = len(rules)

    # calculate fitness (for comparison purposes)
    def calculate_fitness(self, support, confidence):
        return (support + confidence) / 2

    def add_fitness(self):
        for rule in self.rules:
            rule.fitness = self.calculate_fitness(
                rule.support, rule.confidence)

    def sort_rules(self):
        self.rules.sort(key=lambda x: x.fitness, reverse=True)

    def rules_to_csv(self, filename):
        r"""Store rules to CSV file."""
        with open(filename, 'w',) as csvfile:
            writer = csv.writer(csvfile)
            # header of our csv file
            writer.writerow(['Antecedent', 'Consequent',
                            'Fitness', 'Support', 'Confidence'])
            for rule in self.rules:
                writer.writerow([rule.antecedent, rule.consequent,
                                rule.fitness, rule.support, rule.confidence])

    def generate_statistics(self):
        r"""Generate statistics for experimental purposes"""
        fitness = sum(self.calculate_fitness(rule.support, rule.confidence)
                      for rule in self.rules)
        support = sum(rule.support for rule in self.rules)
        confidence = sum(rule.confidence for rule in self.rules)

        print("Total rules: ", self.num_rules)
        print("Average fitness: ", fitness / self.num_rules)
        print("Average support: ", support / self.num_rules)
        print("Average confidence: ", confidence / self.num_rules)

    def generate_stats_report(self, num_rules):
        fitness = [self.calculate_fitness(
            self.rules[i].support, self.rules[i].confidence) for i in range(num_rules)]
        support = [self.rules[i].support for i in range(num_rules)]
        confidence = [self.rules[i].confidence for i in range(num_rules)]

        print("-----------------------------------")
        print(f"Fitness:\n"
              f"Max: {round(max(fitness), 3)}\n"
              f"Min: {round(min(fitness), 3)}\n"
              f"Mean: {round(statistics.mean(fitness), 3)}\n"
              f"Median: {round(statistics.median(fitness), 3)}\n"
              f"Std: {round(statistics.stdev(fitness), 3)}\n")
        print("-----------------------------------")



from niaarm import NiaARM
from niaarm.dataset import Dataset
from niapy.algorithms.basic import DifferentialEvolution
from niapy.task import Task, OptimizationType


class Compare:
    def __init__(self, dataset, output):
        self.dataset = Dataset(dataset)
        self.output = output

    def compare_niaarm(self, np, evaluations):
        self.problem = NiaARM(self.dataset.dimension, self.dataset.features,
                              self.dataset.transactions, metrics=('support', 'confidence'), logging=True)

        task = Task(problem=self.problem, max_evals=evaluations,
                    optimization_type=OptimizationType.MAXIMIZATION)

        algo = DifferentialEvolution(
            population_size=np, differential_weight=0.5, crossover_probability=0.9)

        best = algo.run(task=task)

        self.problem.rules.sort()

        self.problem.rules.to_csv(self.output)

    def get_rules(self):
        return self.problem.rules



from niaarm import Dataset, Feature, Rule
from tinynarm.item import Item
import csv
import sys


class Discretization:
    r"""Main class for discretization tasks.

   Args:
       dataset (csv file): Dataset stored in CSV file.
       num_intervals (int): Number which defines how many intervals we create for numerical features.
   """

    def __init__(self, dataset, num_intervals):
        # load dataset from csv
        self.data = Dataset(dataset)
        self.num_features = len(self.data.features)
        self.num_intervals = num_intervals
        self.feat = []

    def create_intervals(self):
        r"""Create intervals.

        Note: The number of intervals for categorical feature is equal to number of categories.
        """
        for feature in self.data.features:
            if feature.categories is None:
                intervals = self.numerical_interval(
                    feature.min_val, feature.max_val)
                occurences = [0] * self.num_intervals
            else:
                intervals = feature.categories
                occurences = [0] * len(intervals)

            self.feat.append(
                Item(
                    feature.name,
                    feature.dtype,
                    intervals,
                    occurences))

    def numerical_interval(self, min_val, max_val):
        r"""Create intervals for numerical feature."""
        val_range = (max_val - min_val) / self.num_intervals
        intervals = [min_val + (i * val_range)
                     for i in range(self.num_intervals + 1)]
        return intervals

    def generate_dataset(self):
        r"""Create new dataset."""
        self.create_intervals()

        transactions = self.data.transactions.to_numpy()
        discretized_transactions = []

        for transaction in transactions:
            current_transaction = []

            for i, val in enumerate(transaction):
                if self.feat[i].dtype == "cat":
                    current_transaction.append(val)
                else:
                    intervals = self.feat[i].intervals
                    id_interval = 1

                    for j in range(len(intervals) - 1):
                        if intervals[j] <= val < intervals[j+1]:
                            curr = "interval_" + str(id_interval)
                            current_transaction.append(curr)
                            break
                        id_interval += 1
                    else:
                        curr = "interval_" + str(id_interval-1)
                        current_transaction.append(curr)

            discretized_transactions.append(current_transaction)

        return discretized_transactions

    def dataset_to_csv(self, transactions, filename):
        r"""Store dataset to CSV file."""
        with open(filename, 'w', newline='') as csvfile:
            writer = csv.writer(csvfile)
            header = [feature.name for feature in self.data.features]
            writer.writerow(header)
            writer.writerows(transactions)



import numpy as np
from itertools import permutations, combinations
from niaarm import Dataset, Feature, Rule
from tinynarm.item import Item


class TinyNarm:
    r"""Main class for tinyNARM approach.

   Args:
       dataset (csv file): Dataset stored in csv file.
       num_intervals (int): Number which defines how many intervals we create for numerical features.
   """

    def __init__(self, dataset):
        # load dataset from csv
        self.data = Dataset(dataset)
        self.num_features = len(self.data.features)
        self.feat = []
        self.rules = []

    def prepare(self):
        for feature in self.data.features:
            self.feat.append(Item(
                feature.name,
                feature.dtype,
                feature.categories,
                [0] * len(feature.categories)
            ))

    # create item/attribute map
    def calculate_frequencies(self):
        transactions = self.data.transactions.to_numpy()
        for i, transaction in enumerate(transactions):
            for j, val in enumerate(transaction):
                self.feat[j].occurences[self.feat[j].intervals.index(val)] += 1

    def ant_con(self, combination, cut):
        return combination[:cut], combination[cut:]

    def create_rules(self):
        r"""Create new association rules."""

        self.prepare()
        self.calculate_frequencies()

        items = []
        for item in self.feat:
            max_index = np.argmax(item.occurences)
            items.append(Feature(item.name, item.dtype,
                         categories=[item.intervals[max_index]]))

        for i in range(2, self.num_features):
            comb = combinations(items, i)
            if i == 2:
                for j in comb:
                    rule = Rule([j[0]], [j[1]],
                                transactions=self.data.transactions)
                    if rule.support > 0.0:
                        self.rules.append(rule)
            else:
                for j in comb:
                    for cut in range(1, i - 1):
                        ant, con = self.ant_con(j, cut)
                        rule = Rule(
                            ant, con, transactions=self.data.transactions)
                        if rule.support > 0.0:
                            self.rules.append(rule)

    def generate_report(self):
        for f in self.feat:
            print(f"Feat INFO:\n"
                  f"Name: {f.name}\n"
                  f"Bins: {f.intervals}")



from tinynarm.tinynarm import TinyNarm
from tinynarm.item import Item
from tinynarm.utils import Utils
from tinynarm.comparison import Compare

__all__ = [
    "TinyNarm",
    "Item",
    "Discretization",
    "Utils",
    "Compare"
]

__project__ = "tinynarm"
__version__ = '0.2.2'
